# -*- coding: utf-8 -*-
"""a toolset for dealing with polynomials"""
__copyright__ = ("Copyright (c) 2023 https://github.com/dxstiny")


from typing import Optional, Tuple, Union, Set
import sympy as sp


def poles(function: sp.Function, x: sp.Symbol = sp.Symbol("x")) -> Set[sp.Number]:
    """
    Returns the poles of a function.
    :param function: the function
    :return: the poles
    """
    d = sp.solvers.solvers.denoms(function, x)
    ps = set()
    for i in d:
        ps.update(sp.solve(i, x))
    return ps

def zeros(function: sp.Function, x: sp.Symbol = sp.Symbol("x")) -> Set[sp.Number]:
    """
    Returns the zeros of a function.
    :param function: the function
    :return: the zeros
    """
    n, _ = sp.fraction(function)
    return set(sp.Poly(n, x).all_roots())

def isContinuous(function: sp.Function,
                 interval: Optional[Union[sp.Interval,
                                          Tuple[float, float]]] = None) -> bool:
    """
    Checks if a function is continuous.
    :param function: the function
    :param interval: the interval
    """
    if interval is None:
        interval = sp.Interval(-sp.oo, sp.oo)

    if isinstance(interval, tuple):
        interval = sp.Interval(*interval)

    ps = poles(function)

    for pole in ps:
        if pole in interval:
            return False

    return True



# -*- coding: utf-8 -*-
"""functions to determine critical points and extrema"""
__copyright__ = ("Copyright (c) 2023 https://github.com/dxstiny")


from typing import List, Dict
import sympy as sp


def criticalPoints(function: sp.Function,
                   x: sp.Symbol = sp.Symbol("x")) -> List[sp.Number]:
    """Returns the critical points of a function of one variable.
    :param function: function of one variable
    :param x: variable
    :return: critical points of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.criticalPoints import criticalPoints
    >>> x = symbols('x')
    >>> criticalPoints(x**2, x)
    [0]
    """
    return sp.solve(sp.diff(function, x), x, dict=True)

def minima(function: sp.Function,
            x: sp.Symbol = sp.Symbol("x")) -> List[sp.Number]:
    """Returns the minima of a function of one variable.
    :param function: function of one variable
    :param x: variable
    :return: minima of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.criticalPoints import minima
    >>> x = symbols('x')
    >>> minima(x**2, x)
    [0]
    """
    return [ point
             for point in criticalPoints(function, x)
             if sp.diff(function, x, 2).subs(point) > 0 ]

def maxima(function: sp.Function,
            x: sp.Symbol = sp.Symbol("x")) -> List[sp.Number]:
    """Returns the maxima of a function of one variable.
    :param function: function of one variable
    :param x: variable
    :return: maxima of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.criticalPoints import maxima
    >>> x = symbols('x')
    >>> maxima(-x**2, x)
    [0]
    """
    return [ point
             for point in criticalPoints(function, x)
             if sp.diff(function, x, 2).subs(point) < 0]

def turningPoints(function: sp.Function,
                  x: sp.Symbol = sp.Symbol("x")) -> List[sp.Number]:
    """Returns the turning points of a function of one variable.
    :param function: function of one variable
    :param x: variable
    :return: turning points of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.criticalPoints import turningPoints
    >>> x = symbols('x')
    >>> turningPoints(x**2, x)
    [0]
    """
    points: List[Dict[sp.Symbol, sp.Number]] = sp.solve(sp.diff(function, x, 2), x, dict=True)
    return [ point[x]
             for point in points
             if sp.diff(function, x, 3).subs(point) != 0 ]

def saddlePoints(function: sp.Function,
                 x: sp.Symbol = sp.Symbol("x")) -> List[sp.Number]:
    """Returns the saddle points of a function of one variable.
    :param function: function of one variable
    :param x: variable
    :return: saddle points of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.criticalPoints import saddlePoints
    >>> x = symbols('x')
    >>> saddlePoints(x**2, x)
    []
    """
    points = turningPoints(function, x)
    return [ point
             for point in points
             if sp.diff(function, x).subs(x, point) == 0 ]



# -*- coding: utf-8 -*-
"""a toolset for dealing with differentials and differential approximation"""
__copyright__ = ("Copyright (c) 2023 https://github.com/dxstiny")


from typing import Tuple

import sympy as sp


def absoluteDifferential(function: sp.Function,
                         x0: Tuple[sp.Symbol, float],
                         dx: float) -> sp.Function:
    """
    Returns the absolute differential of a function
    :param function: function
    :param x0: point
    :param dx: differential
    :return: differential approximation of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.differential import differentialApproximation
    >>> d = symbols('d')
    >>> V = 1/12 * sp.pi * d ** 3
    >>> differentialApproximation(V, (d, 26), 0.5)
    84.5 * π
    """
    x = x0[0]
    xx = x0[1]
    df = sp.diff(function, x)
    return df.subs(x, xx) * dx


def relativeDifferential(function: sp.Function,
                         x0: Tuple[sp.Symbol, float],
                         dx: float) -> sp.Function:
    """
    Returns the relative differential of a function
    :param function: function
    :param x0: point
    :param dx: differential
    :return: relative differential approximation of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.differential import relativeDifferential
    >>> d = symbols('d')
    >>> V = 1/12 * sp.pi * d ** 3
    >>> relativeDifferential(V, (d, 26), 0.01)
    0.0576923076923077
    """
    return absoluteDifferential(function, x0, dx) / function.subs(x0[0], x0[1])



# -*- coding: utf-8 -*-
"""a toolset for dealing with taylor values & series"""
__copyright__ = ("Copyright (c) 2023 https://github.com/dxstiny")


from typing import Union, Tuple

import sympy as sp
import numpy as np
import matplotlib.pyplot as plt


def TnValue(f: sp.Function, # pylint: disable=invalid-name
            x0: float,
            n: int,
            symbol: sp.Symbol = sp.Symbol("x")) -> sp.Function:
    """
    Returns the nth Taylor value of f(x) at x0.
    :param f: function f(x)
    :param x0: point x0
    :param n: nth Taylor value
    :param symbol: symbol of x
    :return: nth Taylor value of f(x) at x0
    """
    df = sp.diff(f, symbol, n).subs(symbol, x0)
    return df / sp.factorial(n) * (symbol - x0)**n

def Tn(f: sp.Function, # pylint: disable=invalid-name
       x0: float,
       n: int,
       symbol: sp.Symbol = sp.Symbol("x")) -> sp.Function:
    """
    Returns the nth Taylor polynomial of f(x) at x0.
    :param f: function f(x)
    :param x0: point x0
    :param n: nth Taylor polynomial
    :param symbol: symbol of x
    :return: nth Taylor polynomial of f(x) at x0
    """
    tn = 0
    for i in range(n + 1):
        tn += TnValue(f, x0, i, symbol)
    return tn

def TnErrorFunction(f: sp.Function, # pylint: disable=invalid-name
                    x0: float,
                    n: int,
                    symbol: sp.Symbol = sp.Symbol("x")) -> sp.Function:
    """
    Returns the nth Taylor error function of f(x) at x0.
    :param f: function f(x)
    :param x0: point x0
    :param n: nth Taylor error
    :param symbol: symbol of x
    :return: nth Taylor error function of f(x) at x0
    """
    return f - Tn(f, x0, n, symbol)

def TnError(f: sp.Function, # pylint: disable=invalid-name
            x0: float,
            a: float,
            b: float,
            n: int,
            symbol: sp.Symbol = sp.Symbol("x")) -> float:
    """
    Returns the nth Taylor error of f(x) at x0.
    :param f: function f(x)
    :param x0: point x0
    :param a: left bound
    :param b: right bound
    :param n: nth Taylor error
    :param symbol: symbol of x
    :return: nth Taylor error of f(x) at x0
    """
    errf = TnErrorFunction(f, x0, n, symbol)
    xx = np.linspace(a, b, 100)
    errfnp = sp.lambdify(symbol, errf, "numpy")
    yy = errfnp(xx)
    return np.max(np.abs(yy))

def lagrangeRemainder(f: sp.Function,
                      x0: float,
                      a: float,
                      b: float,
                      n: int,
                      symbol: sp.Symbol = sp.Symbol("x")) -> sp.Function:
    """
    Returns the Lagrange remainder of f(x) at x0.
    :param f: function f(x)
    :param x0: point x0
    :param a: left bound
    :param b: right bound
    :param n: nth Lagrange remainder
    :param symbol: symbol of x
    :return: nth Lagrange remainder of f(x) at x0
    """
    n += 1

    x = sp.Symbol("x")
    M = sp.maximum((x - x0)**n, x, sp.Interval(abs(a), abs(b)))
    m = sp.root(M, n) + x0

    dfn = sp.diff(f, symbol, n)

    if dfn.is_negative:
        dfn = -dfn

    g = dfn * M / sp.factorial(n)

    T = sp.maximum(g, symbol, sp.Interval(x0, m))

    if float(T) == 0:
        if n > 2:
            print(f"Warning: Lagrange Remainder is likely incorrect (T = 0) - returning for n={n - 2} instead") # pylint: disable=line-too-long
            return lagrangeRemainder(f, x0, a, b, n - 2, symbol)

        print("Warning: Lagrange Remainder is likely incorrect (T = 0) - returning 0")
        return 0
    return sp.Abs(T)

def TnMin(f: sp.Function, # pylint: disable=invalid-name
          x0: float,
          a: float,
          b: float,
          eps: float,
          symbol: sp.Symbol = sp.Symbol("x")) -> int:
    """
    Returns the minimum n such that the nth Taylor error of f(x) at x0 is less than eps.
    :param f: function f(x)
    :param x0: point x0
    :param a: left bound
    :param b: right bound
    :param eps: maximum error
    :param symbol: symbol of x
    :return: minimum n such that the nth Taylor error of f(x) at x0 is less than eps
    """
    n = 0
    while TnError(f, x0, a, b, n, symbol) > eps:
        n += 1
    return n

def taylorPlot(f: sp.Function,
               x0: float,
               n: int,
               interval: Union[sp.Interval, Tuple[float, float]] = (-1, 1),
               x: sp.Symbol = sp.Symbol("x")) -> None:
    """
    Plots the nth Taylor polynomial *and* the function f(x) on the same graph.
    blue: f(x)
    orange: T_n(x)
    :param f: function f(x)
    :param x0: point x0
    :param n: nth Taylor polynomial
    :param interval: interval to plot on
    :param x: symbol of x
    :return: None
    """
    tn = Tn(f, x0, n, x)

    if isinstance(interval, sp.Interval):
        a, b = interval.args
    else:
        a, b = interval

    xx = np.linspace(x0 + a, x0 + b, 1000)
    fnp = sp.lambdify(x, f, "numpy")
    tnp = sp.lambdify(x, tn, "numpy")
    plt.plot(xx, fnp(xx), label="f(x)")
    plt.plot(xx, tnp(xx), label=f"T_{n}(x)")



# -*- coding: utf-8 -*-
"""general purpose functions for series"""
__copyright__ = ("Copyright (c) 2023 https://github.com/dxstiny")


import matplotlib.pyplot as plt
import sympy as sp

def seriesPlot(series: sp.Expr, n: int) -> None:
    """
    Plot the given series.
    :param series: The series to plot.
    :param n: The number of elements to plot.
    """
    # Determine the elements of the series.
    elements = [series.subs("n", i) for i in range(1, n + 1)]

    # Plot the elements of the series.
    plt.plot(elements, "o-")



# -*- coding: utf-8 -*-
"""a toolset for dealing with geometric sequences & series"""
__copyright__ = ("Copyright (c) 2023 https://github.com/dxstiny")


import sympy as sp

def determine(series: sp.Expr,
              x: sp.Symbol = sp.Symbol("x")) -> sp.Expr:
    """
    Determine a power series.
    :param function: function to determine, must be a power series, with x and k
    :param x: symbol of the function
    :return: series
    """
    series1 = series.subs(x, 1)
    return series1



# -*- coding: utf-8 -*-
"""a toolset for dealing with geometric sequences & series"""
__copyright__ = ("Copyright (c) 2023 https://github.com/dxstiny")


from typing import Dict
import sympy as sp

def fromElements(knownElements: Dict[int, float],
                 counter: sp.Symbol = sp.Symbol("n")) -> sp.Expr:
    """
    Determine the geometric series with the given elements.
    :param knownElements: The known elements of the series. -> {n: a_n}
    :return: The geometric series. -> a_n = a_1 * q ** (n - 1)
    """
    assert len(knownElements) >= 2, "The number of known elements must be at least 2."

    knownIndices = list(knownElements.keys())

    # Determine the common ratio.
    commonRatio = knownElements[knownIndices[1]] / knownElements[knownIndices[0]]
    neg = commonRatio < 0
    commonRatio = abs(commonRatio)
    commonRatio = sp.root(commonRatio, knownIndices[1] - knownIndices[0])
    if neg:
        commonRatio = -commonRatio

    # Determine the first element.
    firstElement = knownElements[knownIndices[0]] / commonRatio ** (knownIndices[0] - 1)

    # Determine the geometric series.
    return firstElement * commonRatio ** (counter - 1)

def sumFromElements(*elements: float) -> sp.Expr:
    """
    Determine the sum of the given elements.
    :param elements: The elements to sum.
    :return: The sum of the given elements.
    """
    series = fromElements({i + 1: elements[i] for i in range(len(elements))})
    return sp.Sum(series, (sp.Symbol("n"), 1, sp.oo)).doit()

def sumFromFirstElement(firstElement: float, q: float) -> sp.Expr:
    """
    Determine the sum of the first n elements of a geometric series.
    :param firstElement: The first element of the series.
    :param commonRatio: The common ratio of the series.
    :return: The sum of all elements of the series.
    """
    return (firstElement / (1 - q)).simplify()



# -*- coding: utf-8 -*-
"""a toolset for dealing with derivatives"""
__copyright__ = ("Copyright (c) 2023 https://github.com/dxstiny")


from typing import Tuple, Optional
import sympy as sp


def newtonRaphson(function: sp.Function,
                  x0: float,
                  n: int = 10,
                  x: sp.Symbol = sp.Symbol("x")) -> float:
    """
    Newton-Raphson method for finding a root of f(x) = 0.
    :param f: function f(x)
    :param x0: initial guess
    :param n: number of iterations
    :param x: variable
    :return: sequence of approximations
    """
    f = sp.lambdify(x, function, "numpy")
    df = sp.lambdify(x, function.diff(x), "numpy")
    x = x0
    for _ in range(n):
        x = x - f(x) / df(x)
    return x

def newtonRaphsonWithEpsilon(function: sp.Function,
                             x0: float,
                             epsilon: float,
                             maxiter: int = 1000,
                             x: sp.Symbol = sp.Symbol("x")) -> Optional[Tuple[int, float]]:
    """
    Newton-Raphson method for finding a root of f(x) = 0.
    :param f: function f(x)
    :param x0: initial guess
    :epsilon: the error tolerance
    :param x: variable
    :return: number of iterations and the root (n, x)
    """
    f = sp.lambdify(x, function, "numpy")
    df = sp.lambdify(x, function.diff(x), "numpy")
    x = x0
    for i in range(maxiter):
        x = x - f(x) / df(x)
        if abs(f(x)) < epsilon:
            return i, x
    return None

def oneSidedDerivative(function: sp.Function,
                       x0: float,
                       direction: str = "+",
                       x: sp.Symbol = sp.Symbol("x")) -> sp.Number:
    """
    One-sided derivative
    :param f: function f(x)
    :param x0: point at which to approximate derivative
    :param direction: direction of approximation ("+" or "-")
    :param x: variable
    :return: derivative
    """
    dx = sp.Symbol("dx")
    nom = function.subs(x, x0 + dx) - function.subs(x, x0)
    denom = dx
    return sp.limit(nom / denom, dx, 0, direction)

def existsDerivative(function: sp.Function,
                     x0: float,
                     x: sp.Symbol = sp.Symbol("x")) -> bool:
    """
    Check if a derivative exists at a point
    :param f: function f(x)
    :param x0: point at which to check
    :param x: variable
    :return: True if derivative exists, False otherwise
    """
    return oneSidedDerivative(function, x0, "+", x) == oneSidedDerivative(function, x0, "-", x)

def curvature(function: sp.Function,
              x: sp.Symbol = sp.Symbol("x")) -> sp.Expr:
    """
    Curvature at a point
    :param f: function f(x)
    :param x: variable
    :return: curvature
    """
    return function.diff(x, 2) / (1 + function.diff(x) ** 2) ** (3 / 2)

def radiusOfCurvature(function: sp.Function,
                      x: sp.Symbol = sp.Symbol("x")) -> sp.Expr:
    """
    Radius of curvature at a point
    :param f: function f(x)
    :param x: variable
    :return: radius of curvature
    """
    return 1 / curvature(function, x)

def centreOfCurvature(function: sp.Function,
                      x: sp.Symbol = sp.Symbol("x")) -> Tuple[sp.Expr, sp.Expr]:
    """
    Centre of curvature at a point
    :param f: function f(x)
    :param x: variable
    :return: centre of curvature
    """
    df = function.diff(x)
    tmpl = (1 + df ** 2) / function.diff(x, 2)
    mx = x - df * tmpl
    my = function + tmpl
    return mx, my



# -*- coding: utf-8 -*-
"""a toolset for dealing with arithmetic sequences & series"""
__copyright__ = ("Copyright (c) 2023 https://github.com/dxstiny")


from typing import Dict
import sympy as sp

def fromElements(knownElements: Dict[int, float],
                 counter: sp.Symbol = sp.Symbol("n")) -> sp.Expr:
    """Determine the arithmetic series with the given elements.
    :param knownElements: The known elements of the series. -> {n: a_n}
    :return: The arithmetic series. -> a_n = a_1 + (n - 1) * d
    """
    assert len(knownElements) >= 2, "The number of known elements must be at least 2."

    knownIndices = list(knownElements.keys())

    # Determine the common difference.
    commonDifference = (knownElements[knownIndices[1]] - knownElements[knownIndices[0]]) / (knownIndices[1] - knownIndices[0]) # pylint: disable=line-too-long

    # Determine the first element.
    firstElement = knownElements[knownIndices[0]] - (knownIndices[0] - 1) * commonDifference

    # Determine the arithmetic series.
    return firstElement + (counter - 1) * commonDifference



# -*- coding: utf-8 -*-
"""a toolset for dealing with multidimensional differentials and differential approximation"""
__copyright__ = ("Copyright (c) 2023 https://github.com/dxstiny")


from typing import Tuple

import sympy as sp


def totalDifferential(function: sp.Function,
                      differentials: Tuple[sp.Symbol],
                      *variables: sp.Symbol) -> sp.Function:
    """Returns the total differential of a function of any number of variables.
    :param function: function of two or more variables
    :param differentials: differentials
    :param variables: variables
    :return: total differential of function
    """
    assert len(differentials) == len(variables)
    return sum( sp.diff(function, v) * d for v, d in zip(variables, differentials) )

def absoluteDifferential(function: sp.Function,
                         x0: Tuple[float, ...],
                         dx: Tuple[float, ...],
                         *variables: sp.Symbol) -> sp.Function:
    """Returns the absolute differential of a function w/ multiple variables
    :param function: function
    :param variables: variables
    :param x0: point
    :param dx: absolute differential of x0
    :return: absolute differential of function
    """
    assert len(x0) == len(variables)
    assert len(dx) == len(variables)

    derivatives = [ sp.diff(function, v) for v in variables ]
    return sum( d.subs(zip(variables, x0)) * e for d, e in zip(derivatives, dx) )

def relativeDifferential(function: sp.Function,
                         x0: Tuple[float, ...],
                         dx: Tuple[float, ...],
                         *variables: sp.Symbol) -> sp.Function:
    """Returns the relative differential of a function w/ multiple variables
    :param function: function
    :param variables: variables
    :param x0: point
    :param dx: relative differential of x0
    :return: relative differential of function
    """
    return absoluteDifferential(function, x0, dx, *variables) / function.subs(zip(variables, x0))



# -*- coding: utf-8 -*-
"""a toolset for dealing with numerical integral approximation"""
__copyright__ = ("Copyright (c) 2023 https://github.com/dxstiny")


from typing import Union
import numpy as np
import sympy as sp


def riemannSum(function: sp.Function,
               a: float,
               b: float,
               n: int,
               upper: bool = False,
               x: sp.Symbol = sp.Symbol("x")) -> float:
    """
    Riemann sum for approximating the integral of f(x) in [a, b].
    :param f: function f(x)
    :param a: left endpoint of interval
    :param b: right endpoint of interval
    :param n: number of subintervals
    :param upper: if True, use upper (right) Riemann sum
    :return: approximation of integral
    """
    f = sp.lambdify(x, function, "numpy")
    h = (b-a)/n
    xx = np.linspace(a, b, n+1)
    y = f(xx)
    if upper:
        return h * sum(y[1:])
    return h * sum(y[:-1])

def trapezoidalRule(function: sp.Function,
                    a: float,
                    b: float,
                    n: int,
                    x: sp.Symbol = sp.Symbol("x")) -> float:
    """
    Trapezoidal rule for approximating the integral of f(x) in [a, b].
    :param f: function f(x)
    :param a: left endpoint of interval
    :param b: right endpoint of interval
    :param n: number of subintervals
    :return: approximation of integral
    """
    f = sp.lambdify(x, function, "numpy")
    h = (b-a)/n
    xx = np.linspace(a, b, n+1)
    y = f(xx)
    return h/2 * (y[0] + 2*sum(y[1:-1]) + y[-1])

def simpsonsRule(function: sp.Function,
                 a: float,
                 b: float,
                 n: int,
                 x: sp.Symbol = sp.Symbol("x")) -> float:
    """
    Simpson's rule for approximating the integral of f(x) in [a, b].
    :param f: function f(x)
    :param a: left endpoint of interval
    :param b: right endpoint of interval
    :param n: number of subintervals
    :return: approximation of integral
    """
    f = sp.lambdify(x, function, "numpy")
    h = (b-a)/n
    xx = np.linspace(a, b, n+1)
    y = f(xx)
    return h/3 * (y[0] + 4*sum(y[1:-1:2]) + 2*sum(y[2:-1:2]) + y[-1])

def arc(function: sp.Function,
        x: sp.Symbol = sp.Symbol("x")) -> sp.Function:
    """
    Arc length function for approximating the integral of sqrt(1 + f'(x)^2) in [a, b].
    :param f: function f(x)
    :param x: variable
    :return: arc length function
    """
    df = function.diff(x)
    return sp.sqrt(1 + df**2)

def arcLength(function: sp.Function,
              a: float,
              b: float,
              x: sp.Symbol = sp.Symbol("x")) -> float:
    """
    Arc length for approximating the integral of sqrt(1 + f'(x)^2) in [a, b].
    :param f: function f(x)
    :param a: left endpoint of interval
    :param b: right endpoint of interval
    :param n: number of subintervals
    :return: approximation of integral
    """
    return sp.integrate(arc(function, x), (x, a, b))

def archimedeanSpiralLength(phi1: float,
                            phi2: float,
                            r: Union[sp.Expr, float] = sp.Symbol("phi")) -> float:
    """
    Archimedean spiral length in [phi1, phi2].
    :param phi1: left endpoint of interval
    :param phi2: right endpoint of interval
    :param r: radius
    :return: length
    """
    phi = sp.Symbol("phi")

    if isinstance(r, sp.Expr):
        dr = r.diff(phi)
    else:
        dr = 0

    g = sp.sqrt(r**2 + dr**2)
    return sp.integrate(g, (phi, phi1, phi2))



# -*- coding: utf-8 -*-
"""functions to determine critical points and extrema of multidimensional functions"""
__copyright__ = ("Copyright (c) 2023 https://github.com/dxstiny")


from typing import List, Dict
import sympy as sp
from anlis.multidimensional.derivative import gradient, determinant


def criticalPoints(function: sp.Function,
                   *variables: sp.Symbol) -> List[Dict[sp.Symbol, sp.Number]]:
    """Returns the critical points of a function of any number of variables.
    :param function: function of two or more variables
    :param variables: variables
    :return: critical points of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.criticalPoints import criticalPoints
    >>> x, y = symbols('x y')
    >>> criticalPoints(x**2 + y**2, x, y)
    [{x: 0, y: 0}]
    """
    points = [ ]
    g = gradient(function, *variables)
    points.extend(sp.solve(g, *variables, dict=True))

    if "sqrt" in str(g):
        print("Warning: sqrt in gradient. There may be more critical points.")
        print("Critical points are where the gradient is zero or undefined.")
        print("Gradient: ", g)

    denoms = sp.solvers.solvers.denoms(g, *variables)
    for denom in denoms:
        for sol in sp.solve(denom, *variables, dict=True):
            for var in variables:
                if var not in sol:
                    continue
                if not sol[var].is_real:
                    break
            else:
                points.append(sol)

    return points

def minima(function: sp.Function,
           *variables: sp.Symbol) -> List[Dict[sp.Symbol, sp.Number]]:
    """Returns the minima of a function of any number of variables.
    :param function: function of two or more variables
    :param variables: variables
    :return: minima of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.criticalPoints import minima
    >>> x, y = symbols('x y')
    >>> minima(x**2 + y**2, x, y)
    [{x: 0, y: 0}]
    """
    return [ point
             for point in criticalPoints(function, *variables)
             if determinant(function, list(variables)).subs(point) > 0 ]

def maxima(function: sp.Function,
           *variables: sp.Symbol) -> List[Dict[sp.Symbol, sp.Number]]:
    """Returns the maxima of a function of any number of variables.
    :param function: function of two or more variables
    :param variables: variables
    :return: maxima of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.criticalPoints import maxima
    >>> x, y = symbols('x y')
    >>> maxima(x**2 + y**2, x, y)
    []
    """
    return [ point
             for point in criticalPoints(function, *variables)
             if determinant(function, list(variables)).subs(point) < 0 ]
